// gcc -o search_tool search_tool.c -D_FILE_OFFSET_BITS=64

// Printing the file with decimal offsets:
// xxd -c 16 -g 1 yourfile | awk '{print strtonum("0x" $1), substr($0, 10)}'

/*
__________                   _____                 .__   __     __                                                    
\______   \_______   ____  _/ ____\_____     ____  |__|_/  |_ _/  |_   ____                                           
 |     ___/\_  __ \ /  _ \ \   __\ \__  \   /    \ |  |\   __\\   __\ /  _ \                                          
 |    |     |  | \/(  <_> ) |  |    / __ \_|   |  \|  | |  |   |  |  (  <_> )                                         
 |____|     |__|    \____/  |__|   (____  /|___|  /|__| |__|   |__|   \____/                                          
                                        \/      \/                                                                    
__________ .__                                              _________  .__                      __                    
\______   \|  |__  _______ _____     ______  ____           \_   ___ \ |  |__    ____    ____  |  | __  ____  _______ 
 |     ___/|  |  \ \_  __ \\__  \   /  ___/_/ __ \   ______ /    \  \/ |  |  \ _/ __ \ _/ ___\ |  |/ /_/ __ \ \_  __ \
 |    |    |   Y  \ |  | \/ / __ \_ \___ \ \  ___/  /_____/ \     \____|   Y  \\  ___/ \  \___ |    < \  ___/  |  | \/
 |____|    |___|  / |__|   (____  //____  > \___  >          \______  /|___|  / \___  > \___  >|__|_ \ \___  > |__|   
                \/              \/      \/      \/                  \/      \/      \/      \/      \/     \/         
*/                                                                                                                      
                                                                                                                      
/*
Overview of the Binary Search Tool

This tool is a console-based application written in C designed to efficiently search for lines from one file (needles) in another (haystack). Its primary feature is the ability to handle large files (e.g., 4+ GB) without loading them entirely into memory, making it ideal for resource-constrained systems.
Highlights of the Tool

    Purpose:
        Finds each line (needle) from one file in another sorted file (haystack) using binary search.

    Memory Efficiency:
        Handles files without loading them into memory.
        Uses file operations and minimal dynamic memory allocation for efficient processing.

    Binary Search for Speed:
        Leverages the binary search algorithm for O(log n) search complexity.
        Optimized for sorted files to minimize disk I/O.

    Dynamic Line Handling:
        Reads lines dynamically using a custom function (read_line) that supports lines of non-fixed length.
        Avoids buffer overflows by allocating memory as needed.

    Seamless File Traversal:
        Includes robust logic to find the start of the current line during binary search using file pointers (fseek/ftell).
        Ensures accurate positioning even with varying line lengths or empty lines.

    Supports Large Files:
        Uses 64-bit file offsets (off_t) to work with files exceeding 4 GB.
        Ensures compatibility with large datasets on modern file systems.

    Helpful Debugging Features:
        Includes debug prints (e.g., current position, line comparisons) for easy identification of issues.
        Tracks and prints the last checked line if a match isn't found.

    No External Dependencies:
        Implements all required functionality without relying on external libraries like getline.
        Portable and straightforward to compile (gcc -D_FILE_OFFSET_BITS=64).

    Output:
        For each needle:
            Prints "Found" if the line exists in the haystack.
            Prints "Not Found" and the nearest similar line if the line doesn't exist.

    Edge Cases Covered:
        Handles:
            Empty files.
            Files with different line endings (\n or \r\n).
            Needles longer than typical lines.

Example Use Case

Command:

$ ./search_tool needles.txt haystack.txt

Output:

Needle 'banana' not found. Last checked line: 'apricot'.
Found: 'apple'.

This tool is well-suited for large-scale log processing, sorted dataset searches, or any task requiring efficient string lookup in massive text files.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Replace ftell and fseek with ftello and fseeko
#define ftell ftello
#define fseek fseeko

// reverse:  reverse string s in place
void reverse(char s[])
{
     int i, j;
     char c;

     for (i = 0, j = strlen(s)-1; i<j; i++, j--) {
         c = s[i];
         s[i] = s[j];
         s[j] = c;
     }
}  

// itoa:  convert n to characters in s
void my_itoa(int n, char s[])
{
     int i, sign;

     if ((sign = n) < 0)  /* record sign */
         n = -n;          /* make n positive */
     i = 0;
     do {       /* generate digits in reverse order */
         s[i++] = n % 10 + '0';   /* get next digit */
     } while ((n /= 10) > 0);     /* delete it */
     if (sign < 0)
         s[i++] = '-';
     s[i] = '\0';
     reverse(s);
}  

// Function to read a single line of any length dynamically
char* read_line(FILE* file) {
    size_t size = 0;
    size_t capacity = 128;
    char* line = malloc(capacity);

    if (!line) {
        perror("Memory allocation error");
        exit(EXIT_FAILURE);
    }

    int ch;
    while ((ch = fgetc(file)) != EOF && ch != '\n') {
        if (size + 1 >= capacity) {
            capacity *= 2;
            char* temp = realloc(line, capacity);
            if (!temp) {
                perror("Memory allocation error");
                free(line);
                exit(EXIT_FAILURE);
            }
            line = temp;
        }
        line[size++] = ch;
    }

    if (size == 0 && ch == EOF) {
        free(line);
        return NULL;
    }

    line[size] = '\0';

    // Remove trailing '\r' if present
    if (size > 0 && line[size - 1] == '\r') {
        line[size - 1] = '\0';
    }

    return line;
}

int binary_search_in_file(const char* haystack_filename, const char* needle, char** LastCheckedLine) {

char s[80];

    FILE* file = fopen(haystack_filename, "r");
    if (!file) {
        perror("Error opening haystack file");
        exit(EXIT_FAILURE);
    }

    off_t left = 0, right, mid, current_pos;
    int ch;

    // Initialize LastCheckedLine
    *LastCheckedLine = NULL;

    // Determine the size of the file
    fseek(file, 0, SEEK_END);
    right = ftell(file);

    while (left <= right) {
        mid = left + (right - left) / 2;
        fseek(file, mid, SEEK_SET);

//dbg [
//printf("Checking position: %lld\n",mid);
//dbg ]

// Move to the beginning of the current line
if (mid != 0) {
    do {
        fseeko(file, -1, SEEK_CUR); // Step back by one byte
        ch = fgetc(file);           // Read the current character

//my_itoa(ch,s);
//printf("\n%d=%s",ch,s);
//dbg [
//printf("%c",ch);
//dbg ]

        if (ftell(file) == 1 && ch != '\n') {
            fseeko(file, -1, SEEK_CUR); // If at the start of the file, rewind
            break;
        }
if (ch!='\n') 
        fseeko(file, -1, SEEK_CUR); // Step back by one byte after checking
    } while (ch != '\n');
}

//dbg [
//printf("\n");
//dbg ]

        current_pos = ftell(file); // Save current position before reading the line

//dbg [
//ch = fgetc(file);           // Read the current character
//fseeko(file, -1, SEEK_CUR); // Step back by one byte after checking
//printf("Checking tuned position, symbol: %lld, %d\n",current_pos,ch);
//dbg ]

        char* line = read_line(file);
        if (!line) {
            break;
        }

        // Free the previous LastCheckedLine
        if (*LastCheckedLine) {
            free(*LastCheckedLine);
        }
        *LastCheckedLine = line; // Store the current line as LastCheckedLine

        int cmp = strcmp(line, needle);
//    cmp > 0: line is greater than needle.
//    cmp == 0: line and needle are equal.
//    cmp < 0: line is less than needle.

        if (cmp == 0) {
            fclose(file);
            return 1; // Found
        } else if (cmp < 0) {
            left = ftell(file);//current_pos; // Use the saved position

//dbg [
//printf("Comparing line vs needle: '%s' vs '%s'\n",line, needle);
//printf("left,right: %lld, %lld\n",left,right);
//dbg ]

        } else {
            right = mid - 1;
        }
    }

    fclose(file);
    return 0; // Not found
}

int main(int argc, char* argv[]) {
    if (argc != 3) {
        fprintf(stderr, "Usage: %s <file-with-needles1.txt> <file-with-needles2.txt>\n", argv[0]);
        return EXIT_FAILURE;
    }

    FILE* needles_file = fopen(argv[1], "r");
    if (!needles_file) {
        perror("Error opening needles file");
        return EXIT_FAILURE;
    }

    char* needle;
    char* last_line = NULL;

    // Process each needle line by line
    while ((needle = read_line(needles_file)) != NULL) {

if (binary_search_in_file(argv[2], needle, &last_line)) {
    //printf("Found: 'needle'\n");
} else {
    printf("Needle_Line '%s', not found, last checked Haystack_Line: '%s'.\n", needle, last_line ? last_line : "None");
}
free(last_line); // Don't forget to free the memory!

/*
        if (binary_search_in_file(argv[2], needle)) {
            printf("Found: '%s'\n", needle);
        } else {
            printf("Not Found: '%s'\n", needle);
        }
        free(needle);
*/
    }

    fclose(needles_file);
    return EXIT_SUCCESS;
}

/*
[sanmayce@djudjeto7 ~]$ LC_ALL=C sort search_tool.c >s
[sanmayce@djudjeto7 ~]$ cat s
























                exit(EXIT_FAILURE);
                free(line);
                fseeko(file, -2, SEEK_CUR);
                perror("Memory allocation error");
            break;
            capacity *= 2;
            char* temp = realloc(line, capacity);
            fclose(file);
            if (!temp) {
            left = ftello(file);
            line = temp;
            printf("Found: '%s'\n", needle);
            printf("Not Found: '%s'\n", needle);
            return 1; // Found
            right = mid - 1;
            while (fgetc(file) != '\n' && ftello(file) > 1) {
            }
            }
        // Move to the beginning of the next line
        char* line = read_line(file);
        exit(EXIT_FAILURE);
        exit(EXIT_FAILURE);
        fprintf(stderr, "Usage: %s <file-with-needles1.txt> <file-with-needles2.txt>\n", argv[0]);
        free(line);
        free(line);
        free(needle);
        fseeko(file, mid, SEEK_SET);
        if (!line) {
        if (binary_search_in_file(argv[2], needle)) {
        if (cmp == 0) {
        if (mid != 0) {
        if (size + 1 >= capacity) {
        int cmp = strcmp(line, needle);
        line[size - 1] = '\0';
        line[size++] = ch;
        mid = left + (right - left) / 2;
        perror("Error opening haystack file");
        perror("Error opening needles file");
        perror("Memory allocation error");
        return EXIT_FAILURE;
        return EXIT_FAILURE;
        return NULL;
        }
        }
        }
        }
        }
        } else if (cmp < 0) {
        } else {
        } else {
    // Determine the size of the file
    // Process each needle line by line
    // Remove trailing '\r' if present
    FILE* file = fopen(haystack_filename, "r");
    FILE* needles_file = fopen(argv[1], "r");
    char* line = malloc(capacity);
    char* needle;
    fclose(file);
    fclose(needles_file);
    fseeko(file, 0, SEEK_END);
    if (!file) {
    if (!line) {
    if (!needles_file) {
    if (argc != 3) {
    if (size == 0 && ch == EOF) {
    if (size > 0 && line[size - 1] == '\r') {
    int ch;
    line[size] = '\0';
    off_t left = 0, right, mid;
    return 0; // Not found
    return EXIT_SUCCESS;
    return line;
    right = ftello(file);
    size_t capacity = 128;
    size_t size = 0;
    while ((ch = fgetc(file)) != EOF && ch != '\n') {
    while ((needle = read_line(needles_file)) != NULL) {
    while (left <= right) {
    }
    }
    }
    }
    }
    }
    }
    }
    }
#define fseek fseeko
#define ftell ftello
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// Function to perform binary search in a file
// Function to read a single line of any length dynamically
// Replace ftell and fseek with ftello and fseeko
// gcc -o search_tool search_tool.c -D_FILE_OFFSET_BITS=64
char* read_line(FILE* file) {
int binary_search_in_file(const char* haystack_filename, const char* needle) {
int main(int argc, char* argv[]) {
}
}
}
[sanmayce@djudjeto7 ~]$ ./search_tool search_tool.c s
Found: '// gcc -o search_tool search_tool.c -D_FILE_OFFSET_BITS=64'
Found: ''
Found: '#include <stdio.h>'
Found: '#include <stdlib.h>'
Found: '#include <string.h>'
Found: ''
Found: '// Replace ftell and fseek with ftello and fseeko'
Found: '#define ftell ftello'
Found: '#define fseek fseeko'
Found: ''
Found: '// Function to read a single line of any length dynamically'
Found: 'char* read_line(FILE* file) {'
Found: '    size_t size = 0;'
Found: '    size_t capacity = 128;'
Found: '    char* line = malloc(capacity);'
Found: ''
Found: '    if (!line) {'
Found: '        perror("Memory allocation error");'
Found: '        exit(EXIT_FAILURE);'
Found: '    }'
Found: ''
Found: '    int ch;'
Found: '    while ((ch = fgetc(file)) != EOF && ch != '\n') {'
Found: '        if (size + 1 >= capacity) {'
Found: '            capacity *= 2;'
Found: '            char* temp = realloc(line, capacity);'
Found: '            if (!temp) {'
Found: '                perror("Memory allocation error");'
Found: '                free(line);'
Found: '                exit(EXIT_FAILURE);'
Found: '            }'
Found: '            line = temp;'
Found: '        }'
Found: '        line[size++] = ch;'
Found: '    }'
Found: ''
Found: '    if (size == 0 && ch == EOF) {'
Found: '        free(line);'
Found: '        return NULL;'
Found: '    }'
Found: ''
Found: '    line[size] = '\0';'
Found: ''
Found: '    // Remove trailing '\r' if present'
Found: '    if (size > 0 && line[size - 1] == '\r') {'
Found: '        line[size - 1] = '\0';'
Found: '    }'
Found: ''
Found: '    return line;'
Found: '}'
Found: ''
Found: '// Function to perform binary search in a file'
Found: 'int binary_search_in_file(const char* haystack_filename, const char* needle) {'
Found: '    FILE* file = fopen(haystack_filename, "r");'
Found: '    if (!file) {'
Found: '        perror("Error opening haystack file");'
Found: '        exit(EXIT_FAILURE);'
Found: '    }'
Found: ''
Found: '    off_t left = 0, right, mid;'
Found: ''
Found: '    // Determine the size of the file'
Found: '    fseeko(file, 0, SEEK_END);'
Found: '    right = ftello(file);'
Found: ''
Found: '    while (left <= right) {'
Found: '        mid = left + (right - left) / 2;'
Found: '        fseeko(file, mid, SEEK_SET);'
Found: ''
Found: '        // Move to the beginning of the next line'
Found: '        if (mid != 0) {'
Found: '            while (fgetc(file) != '\n' && ftello(file) > 1) {'
Found: '                fseeko(file, -2, SEEK_CUR);'
Found: '            }'
Found: '        }'
Found: ''
Found: '        char* line = read_line(file);'
Found: '        if (!line) {'
Found: '            break;'
Found: '        }'
Found: ''
Found: '        int cmp = strcmp(line, needle);'
Found: '        free(line);'
Found: ''
Found: '        if (cmp == 0) {'
Found: '            fclose(file);'
Found: '            return 1; // Found'
Found: '        } else if (cmp < 0) {'
Found: '            left = ftello(file);'
Found: '        } else {'
Found: '            right = mid - 1;'
Found: '        }'
Found: '    }'
Found: ''
Found: '    fclose(file);'
Found: '    return 0; // Not found'
Found: '}'
Found: ''
Found: 'int main(int argc, char* argv[]) {'
Found: '    if (argc != 3) {'
Found: '        fprintf(stderr, "Usage: %s <file-with-needles1.txt> <file-with-needles2.txt>\n", argv[0]);'
Found: '        return EXIT_FAILURE;'
Found: '    }'
Found: ''
Found: '    FILE* needles_file = fopen(argv[1], "r");'
Found: '    if (!needles_file) {'
Found: '        perror("Error opening needles file");'
Found: '        return EXIT_FAILURE;'
Found: '    }'
Found: ''
Found: '    char* needle;'
Found: ''
Found: '    // Process each needle line by line'
Found: '    while ((needle = read_line(needles_file)) != NULL) {'
Found: '        if (binary_search_in_file(argv[2], needle)) {'
Found: '            printf("Found: '%s'\n", needle);'
Found: '        } else {'
Found: '            printf("Not Found: '%s'\n", needle);'
Found: '        }'
Found: '        free(needle);'
Found: '    }'
Found: ''
Found: '    fclose(needles_file);'
Found: '    return EXIT_SUCCESS;'
Found: '}'
Found: ''
[sanmayce@djudjeto7 ~]$ 
*/
